--==============================================================================
-- CERN (BE-CO-HT)
-- Converter board common gateware top-level file
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2014-07-24
--
-- version: 1.0
--
-- description:
--    This module is to be instantiated in all pulse converter board designs,
--    for example the CONV-TTL-BLO or CONV-TTL-RS485. It contains a set of
--    modules common to all these boards, that are configurable via generics to
--    accommodate for the application of each board.
--
-- dependencies:
--    general-cores repository [1]
--
-- references:
--   [1] Board IDs for level conversion circuits
--       www.ohwr.org/projects/conv-common-gw/wiki/Board-id
--   [2] Platform-independent core collection webpage on OHWR,
--       http://www.ohwr.org/projects/general-cores/repository
--   [3] ELMA, Access to board data using SNMP and I2C
--       http://www.ohwr.org/documents/227
--
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2014-07-24   Theodor Stana     File created
--==============================================================================
-- TODO: -
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.genram_pkg.all;

entity conv_common_gw is
  generic
  (
    -- Number of repeater channels
    g_nr_chans            : integer := 6;

    -- Board ID -- 4-letter ASCII string indicating the board ID
    -- see [1] for example
    g_board_id            : std_logic_vector(31 downto 0);
    -- Gateware version
    g_gwvers              : std_logic_vector(7 downto 0);

    -- Generate pulse repetition logic with fixed output pulse width
    g_pgen_fixed_width    : boolean;
    -- Pulse width at pulse generator output (valid with fixed output pulse width)
    g_pgen_pwidth         : natural range 20 to 40 := 24;
    -- Duty cycle divider ratio for pulse generator
    -- output pulse will be limited to 1/g_pgen_duty_cycle_div
    g_pgen_duty_cycle_div : natural := 5;
    -- Pulse generator glitch filter length in number of clk_20_i cycles
    g_pgen_gf_len         : integer := 4;

    -- Bicolor LED controller signals
    g_bicolor_led_columns : integer := 6;
    g_bicolor_led_lines   : integer := 2
  );
  port
  (
    -- Clocks
    clk_20_i                : in  std_logic;
    clk_125_p_i             : in  std_logic;
    clk_125_n_i             : in  std_logic;

    -- Reset output signal, synchronous to 20 MHz clock
    rst_n_o                 : out std_logic;

    -- Glitch filter active-low enable signal
    gf_en_n_i               : in  std_logic;

    -- Channel enable
    global_ch_oen_o         : out std_logic;
    pulse_front_oen_o       : out std_logic;
    pulse_rear_oen_o        : out std_logic;
    inv_oen_o               : out std_logic;

    -- Pulse inputs
    pulse_i                 : in  std_logic_vector(g_nr_chans-1 downto 0);
    pulse_o                 : out std_logic_vector(g_nr_chans-1 downto 0);

    -- Channel leds
    led_pulse_o             : out std_logic_vector(g_nr_chans-1 downto 0);

    -- I2C interface
    scl_i                   : in  std_logic;
    scl_o                   : out std_logic;
    scl_en_o                : out std_logic;
    sda_i                   : in  std_logic;
    sda_o                   : out std_logic;
    sda_en_o                : out std_logic;

    -- VME interface
    vme_ga_i                : in  std_logic_vector(4 downto 0);
    vme_gap_i               : in  std_logic;

    -- SPI interface to on-board flash chip
    flash_cs_n_o            : out std_logic;
    flash_sclk_o            : out std_logic;
    flash_mosi_o            : out std_logic;
    flash_miso_i            : in  std_logic;

    -- I2C LED signals -- conect to a bicolor LED of choice
    -- led_i2c_o pulses four times on I2C transfer
    -- led_i2c_err_o reflects the state of SR.I2C_ERR bit in conv_regs
    led_i2c_o               : out std_logic;
    led_i2c_err_o           : out std_logic;

    -- Switch inputs (for readout from converter status register)
    sw_gp_n_i               : in  std_logic_vector(7 downto 0);
    sw_multicast_n_i        : in  std_logic_vector(3 downto 0);

    -- TTL, INV-TTL and rear-panel channel inputs, for reflection in line status register
    line_ttl_i              : in std_logic_vector(g_nr_chans-1 downto 0);
    line_invttl_i           : in std_logic_vector(3 downto 0);
    line_rear_i             : in std_logic_vector(g_nr_chans-1 downto 0);
    line_rear_fs_i          : in std_logic_vector(g_nr_chans-1 downto 0);

    -- Bicolor LED signals
    bicolor_led_state_i     : in  std_logic_vector(2*g_bicolor_led_columns*g_bicolor_led_lines-1 downto 0);
    bicolor_led_col_o       : out std_logic_vector(g_bicolor_led_columns-1 downto 0);
    bicolor_led_line_o      : out std_logic_vector(g_bicolor_led_lines-1 downto 0);
    bicolor_led_line_oen_o  : out std_logic_vector(g_bicolor_led_lines-1 downto 0)
  );
end entity conv_common_gw;


architecture arch of conv_common_gw is

  --============================================================================
  -- Type declarations
  --============================================================================
  type t_pulse_led_cnt is array(0 to g_nr_chans-1) of unsigned(18 downto 0);
  type t_pulse_cnt is array(0 to g_nr_chans-1) of unsigned(31 downto 0);
  type t_ch_pcr is array(0 to g_nr_chans-1) of std_logic_vector(31 downto 0);

  --============================================================================
  -- Constant declarations
  --============================================================================
  -- Number of Wishbone masters and slaves, for wb_crossbar
  constant c_nr_masters      : natural :=  1;
  constant c_nr_slaves       : natural :=  2;

  -- slave order definitions
  constant c_slv_conv_regs    : natural := 0;
  constant c_slv_multiboot    : natural := 1;
  constant c_slv_onewire_mst  : natural := 2;

  -- Converter board registers SDB definition
  constant c_conv_regs_sdb : t_sdb_device := (
    abi_class     => x"0000",              -- undocumented device
    abi_ver_major => x"01",
    abi_ver_minor => x"00",
    wbd_endian    => c_sdb_endian_big,
    wbd_width     => x"7",                 -- 8/16/32-bit port granularity
    sdb_component => (
      addr_first  => x"0000000000000000",
      addr_last   => x"00000000000000ff",
      product     => (
        vendor_id => x"000000000000CE42",  -- CERN
        device_id => x"3ab464e8",          -- echo "conv_regs          " | md5sum | cut -c1-8
        version   => x"00000001",
        date      => x"20140731",
        name      => "conv_regs          ")));

  -- base address definitions
  constant c_addr_conv_regs   : t_wishbone_address := x"00000000";
  constant c_addr_multiboot   : t_wishbone_address := x"00000100";
  constant c_addr_onewire_mst : t_wishbone_address := x"000000c0";
  constant c_addr_sdb         : t_wishbone_address := x"00000200";

--  -- address mask definitions
--  constant c_mask_conv_regs   : t_wishbone_address := x"00000fc0";
--  constant c_mask_multiboot   : t_wishbone_address := x"00000fc0";
--  constant c_mask_onewire_mst : t_wishbone_address := x"00000fe0";
--
--  -- addresses constant for Wishbone crossbar
--  constant c_addresses : t_wishbone_address_array(c_nr_slaves-1 downto 0)
--                  := (
--                       c_slv_conv_regs   => c_addr_conv_regs
--                  --     c_slv_multiboot   => c_addr_multiboot,
--                  --     c_slv_onewire_mst => c_addr_onewire_mst
--                     );
--
--  -- masks constant for Wishbone crossbar
--  constant c_masks : t_wishbone_address_array(c_nr_slaves-1 downto 0)
--                  := (
--                       c_slv_conv_regs   => c_mask_conv_regs
--                  --     c_slv_multiboot   => c_mask_multiboot,
--                  --     c_slv_onewire_mst => c_mask_onewire_mst
--                     );


  -- SDB interconnect layout
  constant c_sdb_layout : t_sdb_record_array(c_nr_slaves-1 downto 0) := (
      c_slv_conv_regs => f_sdb_embed_device(c_conv_regs_sdb, c_addr_conv_regs),
      c_slv_multiboot => f_sdb_embed_device(c_xwb_xil_multiboot_sdb, c_addr_multiboot)
    );

  --============================================================================
  -- Component declarations
  --============================================================================
  ------------------------------------------------------------------------------
  -- Fixed-width reset generator
  ------------------------------------------------------------------------------
  component conv_reset_gen is
    generic
    (
      -- Reset time in number of clk_i cycles
      g_reset_time : positive := 2_000_000
    );
    port
    (
      clk_i   : in  std_logic;
      rst_i   : in  std_logic;
      rst_n_o : out std_logic
    );
  end component conv_reset_gen;

  component conv_pulse_gen is
    generic
    (
      -- This generic enables elaboration of the fixed pulse width logic
      g_with_fixed_pwidth : boolean;

      -- Pulse width, in number of clk_i cycles
      -- Default pulse width (20 MHz clock): 1.2 us
      -- Minimum allowable pulse width (20 MHz clock): 1 us
      -- Maximum allowable pulse width (20 MHz clock): 2 us
      g_pwidth : natural range 20 to 40 := 24;

      -- Duty cycle divider: D = 1/g_duty_cycle_div
      g_duty_cycle_div : natural := 5
    );
    port
    (
      -- Clock and active-low reset inputs
      clk_i         : in  std_logic;
      rst_n_i       : in  std_logic;

      -- Glitch filter enable input
      -- '1' - Glitch filter disabled (glitch-sensitive, no output jitter)
      -- '0' - Glitch filter enabled (glitch-insensitive, with output jitter)
      gf_en_n_i     : in  std_logic;

      -- Enable input, pulse generation is enabled when '1'
      en_i          : in  std_logic;

      -- Trigger input, has to be '1' to assure pulse output with delay no greater
      -- than internal gate delays.
      trig_a_i      : in  std_logic;

      -- Pulse error output, pulses high for one clock cycle when a pulse arrives
      -- within a pulse period
      pulse_err_p_o : out std_logic;

      -- Pulse output, active-high
      -- latency:
      --    glitch filter disabled: none
      --    glitch filter enabled: glitch filter length + 5 clk_i cycles
      pulse_o       : out std_logic
    );
  end component conv_pulse_gen;

  ------------------------------------------------------------------------------
  -- Converter board control registers
  ------------------------------------------------------------------------------
  component conv_regs is
    port (
      rst_n_i                                  : in     std_logic;
      clk_sys_i                                : in     std_logic;
      wb_adr_i                                 : in     std_logic_vector(5 downto 0);
      wb_dat_i                                 : in     std_logic_vector(31 downto 0);
      wb_dat_o                                 : out    std_logic_vector(31 downto 0);
      wb_cyc_i                                 : in     std_logic;
      wb_sel_i                                 : in     std_logic_vector(3 downto 0);
      wb_stb_i                                 : in     std_logic;
      wb_we_i                                  : in     std_logic;
      wb_ack_o                                 : out    std_logic;
      wb_stall_o                               : out    std_logic;
  -- Port for std_logic_vector field: 'ID register bits' in reg: 'BIDR'
      reg_bidr_i                               : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Gateware version' in reg: 'SR'
      reg_sr_gwvers_i                          : in     std_logic_vector(7 downto 0);
  -- Port for std_logic_vector field: 'Status of on-board switches' in reg: 'SR'
      reg_sr_switches_i                        : in     std_logic_vector(7 downto 0);
  -- Port for std_logic_vector field: 'RTM detection lines~\cite{rtm-det}' in reg: 'SR'
      reg_sr_rtm_i                             : in     std_logic_vector(5 downto 0);
  -- Ports for BIT field: 'I2C communication watchdog timeout error' in reg: 'SR'
      reg_sr_i2c_wdto_o                        : out    std_logic;
      reg_sr_i2c_wdto_i                        : in     std_logic;
      reg_sr_i2c_wdto_load_o                   : out    std_logic;
  -- Port for BIT field: 'White Rabbit present' in reg: 'SR'
      reg_sr_wrpres_i                          : in     std_logic;
  -- Ports for BIT field: 'I2C communication error' in reg: 'SR'
      reg_sr_i2c_err_o                         : out    std_logic;
      reg_sr_i2c_err_i                         : in     std_logic;
      reg_sr_i2c_err_load_o                    : out    std_logic;
  -- Ports for BIT field: 'Pulse missed error' in reg: 'SR'
      reg_sr_pmisse_o                          : out    std_logic;
      reg_sr_pmisse_i                          : in     std_logic;
      reg_sr_pmisse_load_o                     : out    std_logic;
  -- Ports for BIT field: 'Reset unlock bit' in reg: 'CR'
      reg_cr_rst_unlock_o                      : out    std_logic;
      reg_cr_rst_unlock_i                      : in     std_logic;
      reg_cr_rst_unlock_load_o                 : out    std_logic;
  -- Ports for BIT field: 'Reset bit' in reg: 'CR'
      reg_cr_rst_o                             : out    std_logic;
      reg_cr_rst_i                             : in     std_logic;
      reg_cr_rst_load_o                        : out    std_logic;
  -- Ports for PASS_THROUGH field: 'Manual Pulse Trigger' in reg: 'CR'
      reg_cr_mpt_o                             : out    std_logic_vector(7 downto 0);
      reg_cr_mpt_wr_o                          : out    std_logic;
  -- Port for std_logic_vector field: 'Pulse counter value' in reg: 'CH1PCR'
      reg_ch1pcr_o                             : out    std_logic_vector(31 downto 0);
      reg_ch1pcr_i                             : in     std_logic_vector(31 downto 0);
      reg_ch1pcr_load_o                        : out    std_logic;
  -- Port for std_logic_vector field: 'Pulse counter value' in reg: 'CH2PCR'
      reg_ch2pcr_o                             : out    std_logic_vector(31 downto 0);
      reg_ch2pcr_i                             : in     std_logic_vector(31 downto 0);
      reg_ch2pcr_load_o                        : out    std_logic;
  -- Port for std_logic_vector field: 'Pulse counter value' in reg: 'CH3PCR'
      reg_ch3pcr_o                             : out    std_logic_vector(31 downto 0);
      reg_ch3pcr_i                             : in     std_logic_vector(31 downto 0);
      reg_ch3pcr_load_o                        : out    std_logic;
  -- Port for std_logic_vector field: 'Pulse counter value' in reg: 'CH4PCR'
      reg_ch4pcr_o                             : out    std_logic_vector(31 downto 0);
      reg_ch4pcr_i                             : in     std_logic_vector(31 downto 0);
      reg_ch4pcr_load_o                        : out    std_logic;
  -- Port for std_logic_vector field: 'Pulse counter value' in reg: 'CH5PCR'
      reg_ch5pcr_o                             : out    std_logic_vector(31 downto 0);
      reg_ch5pcr_i                             : in     std_logic_vector(31 downto 0);
      reg_ch5pcr_load_o                        : out    std_logic;
  -- Port for std_logic_vector field: 'Pulse counter value' in reg: 'CH6PCR'
      reg_ch6pcr_o                             : out    std_logic_vector(31 downto 0);
      reg_ch6pcr_i                             : in     std_logic_vector(31 downto 0);
      reg_ch6pcr_load_o                        : out    std_logic;
  -- Port for std_logic_vector field: 'TAI seconds counter bits 31..0' in reg: 'TVLR'
      reg_tvlr_o                               : out    std_logic_vector(31 downto 0);
      reg_tvlr_i                               : in     std_logic_vector(31 downto 0);
      reg_tvlr_load_o                          : out    std_logic;
  -- Port for std_logic_vector field: 'TAI seconds counter bits 39..32' in reg: 'TVHR'
      reg_tvhr_o                               : out    std_logic_vector(7 downto 0);
      reg_tvhr_i                               : in     std_logic_vector(7 downto 0);
      reg_tvhr_load_o                          : out    std_logic;
  -- Port for std_logic_vector field: 'Channel mask' in reg: 'TBMR'
      reg_tbmr_chan_i                          : in     std_logic_vector(5 downto 0);
  -- Port for BIT field: 'White Rabbit present' in reg: 'TBMR'
      reg_tbmr_wrtag_i                         : in     std_logic;
  -- Tag buffer read request, asserted when reading from TBMR
      reg_tb_rd_req_p_o                        : out    std_logic;
  -- Port for std_logic_vector field: 'Cycles counter' in reg: 'TBCYR'
      reg_tbcyr_i                              : in     std_logic_vector(27 downto 0);
  -- Port for std_logic_vector field: 'Lower part of TAI seconds counter' in reg: 'TBTLR'
      reg_tbtlr_i                              : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Upper part of TAI seconds counter' in reg: 'TBTHR'
      reg_tbthr_i                              : in     std_logic_vector(7 downto 0);
  -- Port for std_logic_vector field: 'Buffer counter' in reg: 'TBCSR'
      reg_tbcsr_usedw_i                        : in     std_logic_vector(6 downto 0);
  -- Port for BIT field: 'Buffer full' in reg: 'TBCSR'
      reg_tbcsr_full_i                         : in     std_logic;
  -- Port for BIT field: 'Buffer empty' in reg: 'TBCSR'
      reg_tbcsr_empty_i                        : in     std_logic;
  -- Ports for BIT field: 'Clear tag buffer' in reg: 'TBCSR'
      reg_tbcsr_clr_o                          : out    std_logic;
      reg_tbcsr_clr_i                          : in     std_logic;
      reg_tbcsr_clr_load_o                     : out    std_logic;
  -- Port for std_logic_vector field: 'Lower part of TAI seconds counter' in reg: 'CH1LTSTLR'
      reg_ch1ltstlr_i                          : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Upper part of TAI seconds counter' in reg: 'CH1LTSTHR'
      reg_ch1ltsthr_tai_i                      : in     std_logic_vector(7 downto 0);
  -- Port for BIT field: 'White Rabbit present' in reg: 'CH1LTSTHR'
      reg_ch1ltsthr_wrtag_i                    : in     std_logic;
  -- Port for std_logic_vector field: 'Lower part of TAI seconds counter' in reg: 'CH2LTSTLR'
      reg_ch2ltstlr_i                          : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Upper part of TAI seconds counter' in reg: 'CH2LTSTHR'
      reg_ch2ltsthr_tai_i                      : in     std_logic_vector(7 downto 0);
  -- Port for BIT field: 'White Rabbit present' in reg: 'CH2LTSTHR'
      reg_ch2ltsthr_wrtag_i                    : in     std_logic;
  -- Port for std_logic_vector field: 'Lower part of TAI seconds counter' in reg: 'CH3LTSTLR'
      reg_ch3ltstlr_i                          : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Upper part of TAI seconds counter' in reg: 'CH3LTSTHR'
      reg_ch3ltsthr_tai_i                      : in     std_logic_vector(7 downto 0);
  -- Port for BIT field: 'White Rabbit present' in reg: 'CH3LTSTHR'
      reg_ch3ltsthr_wrtag_i                    : in     std_logic;
  -- Port for std_logic_vector field: 'Lower part of TAI seconds counter' in reg: 'CH4LTSTLR'
      reg_ch4ltstlr_i                          : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Upper part of TAI seconds counter' in reg: 'CH4LTSTHR'
      reg_ch4ltsthr_tai_i                      : in     std_logic_vector(7 downto 0);
  -- Port for BIT field: 'White Rabbit present' in reg: 'CH4LTSTHR'
      reg_ch4ltsthr_wrtag_i                    : in     std_logic;
  -- Port for std_logic_vector field: 'Lower part of TAI seconds counter' in reg: 'CH5LTSTLR'
      reg_ch5ltstlr_i                          : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Upper part of TAI seconds counter' in reg: 'CH5LTSTHR'
      reg_ch5ltsthr_tai_i                      : in     std_logic_vector(7 downto 0);
  -- Port for BIT field: 'White Rabbit present' in reg: 'CH5LTSTHR'
      reg_ch5ltsthr_wrtag_i                    : in     std_logic;
  -- Port for std_logic_vector field: 'Lower part of TAI seconds counter' in reg: 'CH6LTSTLR'
      reg_ch6ltstlr_i                          : in     std_logic_vector(31 downto 0);
  -- Port for std_logic_vector field: 'Upper part of TAI seconds counter' in reg: 'CH6LTSTHR'
      reg_ch6ltsthr_tai_i                      : in     std_logic_vector(7 downto 0);
  -- Port for BIT field: 'White Rabbit present' in reg: 'CH6LTSTHR'
      reg_ch6ltsthr_wrtag_i                    : in     std_logic;
  -- Port for std_logic_vector field: 'Front panel TTL input state' in reg: 'LSR'
      reg_lsr_front_ttl_i                      : in     std_logic_vector(5 downto 0);
  -- Port for std_logic_vector field: 'Front panel TTL-BAR input state' in reg: 'LSR'
      reg_lsr_front_invttl_i                   : in     std_logic_vector(3 downto 0);
  -- Port for std_logic_vector field: 'Rear panel input state' in reg: 'LSR'
      reg_lsr_rear_i                           : in     std_logic_vector(5 downto 0);
  -- Port for std_logic_vector field: 'Input failsafe state' in reg: 'LSR'
      reg_lsr_fs_i                             : in     std_logic_vector(5 downto 0);
  -- Port for std_logic_vector field: 'State of MultiCast switches' in reg: 'MSWR'
      reg_mswr_bit_i                           : in     std_logic_vector(3 downto 0)
    );
  end component conv_regs;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Per-domain clock and reset signals
  signal clk_125             : std_logic;
  signal rst_125_n           : std_logic;
  signal rst_20_n            : std_logic;

  -- Pulse logic signals
  signal trig_a              : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_synced         : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_degl           : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_chan           : std_logic_vector(g_nr_chans-1 downto 0);
  signal trig_pgen           : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp          : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp_d0       : std_logic_vector(g_nr_chans-1 downto 0);
  signal pulse_outp_redge_p  : std_logic_vector(g_nr_chans-1 downto 0);

  -- I2C bridge signals
  signal i2c_addr            : std_logic_vector(6 downto 0);
  signal i2c_tip             : std_logic;
  signal i2c_err_p           : std_logic;
  signal i2c_wdto_p          : std_logic;

  -- Signals to/from converter system registers component
  signal rtm_lines           : std_logic_vector(5 downto 0);
  signal sw_gp               : std_logic_vector(7 downto 0);
  signal sw_multicast        : std_logic_vector(3 downto 0);
  signal rst_unlock          : std_logic;
  signal rst_unlock_bit      : std_logic;
  signal rst_unlock_bit_ld   : std_logic;
  signal rst_bit             : std_logic;
  signal rst_bit_ld          : std_logic;
  signal rst_fr_reg          : std_logic;
  signal i2c_wdto_bit        : std_logic;
  signal i2c_wdto_bit_rst    : std_logic;
  signal i2c_wdto_bit_rst_ld : std_logic;
  signal pmisse_bit          : std_logic;
  signal pmisse_bit_rst      : std_logic;
  signal pmisse_bit_rst_ld   : std_logic;
  signal pulse_cnt           : t_pulse_cnt;
  signal ch_pcr              : t_ch_pcr;
  signal ch_pcr_ld           : std_logic_vector(5 downto 0);
  signal mpt_ld              : std_logic;
  signal mpt                 : std_logic_vector( 7 downto 0);
  signal tvlr                : std_logic_vector(31 downto 0);
  signal tvlr_ld             : std_logic;
  signal tvhr                : std_logic_vector( 7 downto 0);
  signal tvhr_ld             : std_logic;
  signal wrpres              : std_logic;
  signal i2c_err_bit         : std_logic;
  signal i2c_err_bit_rst     : std_logic;
  signal i2c_err_bit_rst_ld  : std_logic;

  -- LED signals
  signal led_pulse           : std_logic_vector(g_nr_chans-1 downto 0);
  signal led_pulse_cnt       : t_pulse_led_cnt;
  signal led_i2c             : std_logic;
  signal led_i2c_clkdiv      : unsigned(18 downto 0);
  signal led_i2c_cnt         : unsigned( 2 downto 0);
  signal led_i2c_blink       : std_logic;

  -- Wishbone crossbar signals
  signal xbar_slave_in       : t_wishbone_slave_in_array  (c_nr_masters-1 downto 0);
  signal xbar_slave_out      : t_wishbone_slave_out_array (c_nr_masters-1 downto 0);
  signal xbar_master_in      : t_wishbone_master_in_array (c_nr_slaves-1 downto 0);
  signal xbar_master_out     : t_wishbone_master_out_array(c_nr_slaves-1 downto 0);

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- Internal and external reset generation
  --============================================================================
  -- Configure reset generator for 100ms power-on reset
  cmp_reset_gen : conv_reset_gen
    generic map
    (
      -- Reset time: 50ns * 2 * (10**6) = 100 ms
      g_reset_time => 2*(10**6)
    )
    port map
    (
      clk_i   => clk_20_i,
      rst_i   => '0',
      rst_n_o => rst_20_n
    );

  -- Output reset signal is synchronous to the 20 MHz clock
  rst_n_o <= rst_20_n;

  --============================================================================
  -- Output enable logic
  --============================================================================
  global_ch_oen_o   <= '1';
  pulse_front_oen_o <= '1';
  inv_oen_o         <= '1';
  pulse_rear_oen_o  <= '1';

  --============================================================================
  -- Pulse repetition logic
  --============================================================================
  trig_a <= pulse_i;

  gen_pulse_chan_logic : for i in 0 to g_nr_chans-1 generate

    -- Synchronize the asynchronous trigger input into the 20 MHz clock
    -- domain before passing it to the glitch filter
    --
    -- Reset value is '1' to avoid pulses being counted by pulse counter on
    -- startup, when the board is in TTL-BAR repetition mode.
    cmp_trig_sync : gc_sync_ffs
      generic map
      (
        g_sync_edge => "positive"
      )
      port map
      (
        clk_i    => clk_20_i,
        rst_n_i  => rst_20_n,
        data_i   => trig_a(i),
        synced_o => trig_synced(i)
      );

    -- Deglitch synchronized trigger signal
    --
    -- Reset value is '1' to avoid pulses being counted by pulse counter on
    -- startup, when the board is in TTL-BAR repetition mode.
    cmp_inp_glitch_filt : gc_glitch_filt
      generic map
      (
        g_len => g_pgen_gf_len
      )
      port map
      (
        clk_i   => clk_20_i,
        rst_n_i => rst_20_n,
        dat_i   => trig_synced(i),
        dat_o   => trig_degl(i)
      );

    -- Now that we have a deglitched signal, generate the MUX to select between
    -- deglitched and direct channel input
    trig_chan(i) <= trig_a(i) when (gf_en_n_i = '1') else
                    trig_degl(i);

    -- The trigger to the pulse generator is either manual OR from the channel input
    trig_pgen(i) <= trig_chan(i);

    -- Instantiate pulse generator block for the channel
    cmp_pulse_gen : conv_pulse_gen
      generic map
      (
        g_with_fixed_pwidth => g_pgen_fixed_width,
        g_pwidth            => g_pgen_pwidth,
        g_duty_cycle_div    => g_pgen_duty_cycle_div
      )
      port map
      (
        clk_i         => clk_20_i,
        rst_n_i       => rst_20_n,

        gf_en_n_i     => gf_en_n_i,

        en_i          => '1',

        trig_a_i      => trig_pgen(i),

        pulse_err_p_o => open,

        pulse_o       => pulse_outp(i)
      );

    -- Process to flash pulse LED when a pulse is output
    -- LED flash length: 26 ms
    p_pulse_led : process (clk_20_i) is
    begin
      if rising_edge(clk_20_i) then
        if (rst_20_n = '0') then
          pulse_outp_d0(i)      <= '0';
          pulse_outp_redge_p(i) <= '0';
          led_pulse_cnt(i)      <= (others => '0');
          led_pulse(i)          <= '0';
        else
          pulse_outp_d0(i)      <= pulse_outp(i);
          pulse_outp_redge_p(i) <= pulse_outp(i) and (not pulse_outp_d0(i));
          case led_pulse(i) is
            when '0' =>
              if (pulse_outp_redge_p(i) = '1') then
                led_pulse(i) <= '1';
              end if;
            when '1' =>
              led_pulse_cnt(i) <= led_pulse_cnt(i) + 1;
              if (led_pulse_cnt(i) = (led_pulse_cnt(i)'range => '1')) then
                led_pulse(i) <= '0';
              end if;
            when others =>
              led_pulse(i) <= '0';
          end case;
        end if;
      end if;
    end process p_pulse_led;

  end generate gen_pulse_chan_logic;

  -- Channel output assignments
  pulse_o     <= pulse_outp;
  led_pulse_o <= led_pulse;

  --============================================================================
  -- I2C bridge logic
  --============================================================================
  -- Set the I2C address signal according to ELMA protocol [2]
  i2c_addr <= "10" & vme_ga_i;

  -- Instantiate I2C bridge component
  --
  -- FSM watchdog timeout timer:
  -- * consider bit period of 30 us
  -- * 10 bits / byte transfer => 300 us
  -- * 40 bytes in one transfer => 12000 us
  -- * clk_i period = 50 ns => g_fsm_wdt = 12000 us / 50 ns = 240000
  -- * multiply by two for extra safety => g_fsm_wdt = 480000
  -- * Time to watchdog timeout: 480000 * 50ns = 24 ms
  cmp_i2c_bridge : wb_i2c_bridge
    generic map
    (
      g_fsm_wdt => 480000
    )
    port map
    (
      -- Clock, reset
      clk_i      => clk_20_i,
      rst_n_i    => rst_20_n,

      -- I2C lines
      scl_i      => scl_i,
      scl_o      => scl_o,
      scl_en_o   => scl_en_o,
      sda_i      => sda_i,
      sda_o      => sda_o,
      sda_en_o   => sda_en_o,

      -- I2C address and status
      i2c_addr_i => i2c_addr,

      -- TIP and ERR outputs
      tip_o      => i2c_tip,
      err_p_o    => i2c_err_p,
      wdto_p_o   => i2c_wdto_p,

      -- Wishbone master signals
      wbm_stb_o  => xbar_slave_in(0).stb,
      wbm_cyc_o  => xbar_slave_in(0).cyc,
      wbm_sel_o  => xbar_slave_in(0).sel,
      wbm_we_o   => xbar_slave_in(0).we,
      wbm_dat_i  => xbar_slave_out(0).dat,
      wbm_dat_o  => xbar_slave_in(0).dat,
      wbm_adr_o  => xbar_slave_in(0).adr,
      wbm_ack_i  => xbar_slave_out(0).ack,
      wbm_rty_i  => xbar_slave_out(0).rty,
      wbm_err_i  => xbar_slave_out(0).err
    );

  -- Process to blink the LED when an I2C transfer is in progress
  -- blinks four times per transfer
  -- blink width : 20 ms
  -- blink period: 40 ms
  p_i2c_blink : process(clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        led_i2c_clkdiv <= (others => '0');
        led_i2c_cnt    <= (others => '0');
        led_i2c        <= '0';
        led_i2c_blink  <= '0';
      else
        case led_i2c_blink is

          when '0' =>
            led_i2c <= '0';
            if (i2c_tip = '1') then
              led_i2c_blink <= '1';
            end if;

          when '1' =>
            led_i2c_clkdiv <= led_i2c_clkdiv + 1;
            if (led_i2c_clkdiv = 399999) then
              led_i2c_clkdiv <= (others => '0');
              led_i2c_cnt    <= led_i2c_cnt + 1;
              led_i2c        <= not led_i2c;
              if (led_i2c_cnt = 7) then
                led_i2c_cnt <= (others => '0');
                led_i2c_blink <= '0';
              end if;
            end if;

          when others =>
            led_i2c_blink <= '0';

        end case;
      end if;
    end if;
  end process p_i2c_blink;

  -- And set the I2C LED Output
  led_i2c_o <= led_i2c;

  -- Register for the I2C_WDTO bit in the SR, cleared by writing a '1'
  p_sr_wdto_bit : process(clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        i2c_wdto_bit <= '0';
      elsif (i2c_wdto_p = '1') then
        i2c_wdto_bit <= '1';
      elsif (i2c_wdto_bit_rst_ld = '1') and (i2c_wdto_bit_rst = '1') then
        i2c_wdto_bit <= '0';
      end if;
    end if;
  end process p_sr_wdto_bit;

  -- Register for the I2C_ERR bit in the SR
  p_i2c_err_led : process(clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        i2c_err_bit <= '0';
      elsif (i2c_err_p = '1') then
        i2c_err_bit <= '1';
      elsif (i2c_err_bit_rst_ld = '1') and (i2c_err_bit_rst = '1') then
        i2c_err_bit <= '0';
      end if;
    end if;
  end process p_i2c_err_led;

  -- Set the I2C err signal for the LED
  led_i2c_err_o <= i2c_err_bit;

  --============================================================================
  -- Instantiation and connection of the main Wishbone crossbar
  --============================================================================
  cmp_wb_crossbar : xwb_sdb_crossbar
    generic map
    (
      g_num_masters => c_nr_masters,
      g_num_slaves  => c_nr_slaves,
      g_registered  => false,
      g_wraparound  => true,
      g_layout      => c_sdb_layout,
      g_sdb_addr    => c_addr_sdb
    )
    port map
    (
      clk_sys_i => clk_20_i,
      rst_n_i   => rst_20_n,
      slave_i   => xbar_slave_in,
      slave_o   => xbar_slave_out,
      master_i  => xbar_master_in,
      master_o  => xbar_master_out
    );

--  cmp_wb_crossbar : xwb_crossbar
--  generic map
--  (
--    g_num_masters => c_nr_masters,
--    g_num_slaves  => c_nr_slaves,
--    g_registered  => false,
--    g_address     => c_addresses,
--    g_mask        => c_masks
--  )
--  port map
--  (
--    clk_sys_i => clk_20_i,
--    rst_n_i   => rst_20_n,
--    slave_i   => xbar_slave_in,
--    slave_o   => xbar_slave_out,
--    master_i  => xbar_master_in,
--    master_o  => xbar_master_out
--  );

  --============================================================================
  -- Converter board registers
  --============================================================================
  -- Make switches readout as '1' for ON, instead of '0'
  sw_gp        <= not sw_gp_n_i;
  sw_multicast <= not sw_multicast_n_i;

  -- Then, instantiate the component
  cmp_conv_regs : conv_regs
    port map (
      rst_n_i                  => rst_20_n,
      clk_sys_i                => clk_20_i,

      wb_adr_i                 => xbar_master_out(c_slv_conv_regs).adr(7 downto 2),
      wb_dat_i                 => xbar_master_out(c_slv_conv_regs).dat,
      wb_dat_o                 => xbar_master_in (c_slv_conv_regs).dat,
      wb_cyc_i                 => xbar_master_out(c_slv_conv_regs).cyc,
      wb_sel_i                 => xbar_master_out(c_slv_conv_regs).sel,
      wb_stb_i                 => xbar_master_out(c_slv_conv_regs).stb,
      wb_we_i                  => xbar_master_out(c_slv_conv_regs).we,
      wb_ack_o                 => xbar_master_in (c_slv_conv_regs).ack,
      wb_stall_o               => xbar_master_in (c_slv_conv_regs).stall,

      reg_bidr_i               => g_board_id,

      reg_sr_gwvers_i          => g_gwvers,
      reg_sr_switches_i        => sw_gp,
      reg_sr_rtm_i             => (others => '0'),
      reg_sr_i2c_wdto_o        => i2c_wdto_bit_rst,
      reg_sr_i2c_wdto_i        => i2c_wdto_bit,
      reg_sr_i2c_wdto_load_o   => i2c_wdto_bit_rst_ld,
      reg_sr_wrpres_i          => '0',
      reg_sr_i2c_err_o         => i2c_err_bit_rst,
      reg_sr_i2c_err_i         => i2c_err_bit,
      reg_sr_i2c_err_load_o    => i2c_err_bit_rst_ld,
      reg_sr_pmisse_o          => open,
      reg_sr_pmisse_i          => '0',
      reg_sr_pmisse_load_o     => open,

      reg_cr_rst_unlock_o      => rst_unlock_bit,
      reg_cr_rst_unlock_i      => rst_unlock,
      reg_cr_rst_unlock_load_o => rst_unlock_bit_ld,
      reg_cr_rst_o             => rst_bit,
      reg_cr_rst_i             => rst_fr_reg,
      reg_cr_rst_load_o        => rst_bit_ld,
      reg_cr_mpt_o             => open,
      reg_cr_mpt_wr_o          => open,

      reg_ch1pcr_o             => ch_pcr(0),
      reg_ch1pcr_i             => std_logic_vector(pulse_cnt(0)),
      reg_ch1pcr_load_o        => ch_pcr_ld(0),
      reg_ch2pcr_o             => ch_pcr(1),
      reg_ch2pcr_i             => std_logic_vector(pulse_cnt(1)),
      reg_ch2pcr_load_o        => ch_pcr_ld(1),
      reg_ch3pcr_o             => ch_pcr(2),
      reg_ch3pcr_i             => std_logic_vector(pulse_cnt(2)),
      reg_ch3pcr_load_o        => ch_pcr_ld(2),
      reg_ch4pcr_o             => ch_pcr(3),
      reg_ch4pcr_i             => std_logic_vector(pulse_cnt(3)),
      reg_ch4pcr_load_o        => ch_pcr_ld(3),
      reg_ch5pcr_o             => ch_pcr(4),
      reg_ch5pcr_i             => std_logic_vector(pulse_cnt(4)),
      reg_ch5pcr_load_o        => ch_pcr_ld(4),
      reg_ch6pcr_o             => ch_pcr(5),
      reg_ch6pcr_i             => std_logic_vector(pulse_cnt(5)),
      reg_ch6pcr_load_o        => ch_pcr_ld(5),

      reg_tvlr_o               => open,
      reg_tvlr_i               => (others => '0'),
      reg_tvlr_load_o          => open,
      reg_tvhr_o               => open,
      reg_tvhr_i               => (others => '0'),
      reg_tvhr_load_o          => open,

      reg_tbmr_chan_i          => (others => '0'),
      reg_tbmr_wrtag_i         => '0',
      reg_tb_rd_req_p_o        => open,
      reg_tbcyr_i              => (others => '0'),
      reg_tbtlr_i              => (others => '0'),
      reg_tbthr_i              => (others => '0'),
      reg_tbcsr_clr_o          => open,
      reg_tbcsr_clr_i          => '0',
      reg_tbcsr_clr_load_o     => open,
      reg_tbcsr_usedw_i        => (others => '0'),
      reg_tbcsr_full_i         => '0',
      reg_tbcsr_empty_i        => '0',

      reg_ch1ltstlr_i          => (others => '0'),
      reg_ch1ltsthr_tai_i      => (others => '0'),
      reg_ch1ltsthr_wrtag_i    => '0',
      reg_ch2ltstlr_i          => (others => '0'),
      reg_ch2ltsthr_tai_i      => (others => '0'),
      reg_ch2ltsthr_wrtag_i    => '0',
      reg_ch3ltstlr_i          => (others => '0'),
      reg_ch3ltsthr_tai_i      => (others => '0'),
      reg_ch3ltsthr_wrtag_i    => '0',
      reg_ch4ltstlr_i          => (others => '0'),
      reg_ch4ltsthr_tai_i      => (others => '0'),
      reg_ch4ltsthr_wrtag_i    => '0',
      reg_ch5ltstlr_i          => (others => '0'),
      reg_ch5ltsthr_tai_i      => (others => '0'),
      reg_ch5ltsthr_wrtag_i    => '0',
      reg_ch6ltstlr_i          => (others => '0'),
      reg_ch6ltsthr_tai_i      => (others => '0'),
      reg_ch6ltsthr_wrtag_i    => '0',

      reg_lsr_front_ttl_i      => line_ttl_i,
      reg_lsr_front_invttl_i   => line_invttl_i,
      reg_lsr_rear_i           => line_rear_i,
      reg_lsr_fs_i             => line_rear_fs_i,

      reg_mswr_bit_i           => sw_multicast
    );

  -- Implement the RST_UNLOCK bit
  p_rst_unlock : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        rst_unlock <= '0';
      elsif (rst_unlock_bit_ld = '1') then
        if (rst_unlock_bit = '1') then
          rst_unlock <= '1';
        else
          rst_unlock <= '0';
        end if;
      end if;
    end if;
  end process p_rst_unlock;

  -- Implement the reset bit register
  -- The register can only be set when the RST_UNLOCK bit is '1'.
  p_rst_fr_reg : process (clk_20_i)
  begin
    if rising_edge(clk_20_i) then
      if (rst_20_n = '0') then
        rst_fr_reg <= '0';
      elsif (rst_bit_ld = '1') and (rst_bit = '1') and (rst_unlock = '1') then
        rst_fr_reg <= '1';
      else
        rst_fr_reg <= '0';
      end if;
    end if;
  end process p_rst_fr_reg;

  --============================================================================
  -- Instantiate Xilinx MultiBoot module
  --============================================================================
  cmp_multiboot : xwb_xil_multiboot
    port map
    (
      clk_i      => clk_20_i,
      rst_n_i    => rst_20_n,
      wbs_i      => xbar_master_out(c_slv_multiboot),
      wbs_o      => xbar_master_in(c_slv_multiboot),
      spi_cs_n_o => flash_cs_n_o,
      spi_sclk_o => flash_sclk_o,
      spi_mosi_o => flash_mosi_o,
      spi_miso_i => flash_miso_i
    );

  --============================================================================
  -- Bicolor LED matrix logic
  --============================================================================
  -- connect column, line & state signals outside
  cmp_bicolor_led_ctrl : gc_bicolor_led_ctrl
    generic map
    (
      g_NB_COLUMN    => g_bicolor_led_columns,
      g_NB_LINE      => g_bicolor_led_lines,
      g_clk_freq     => 20000000,
      g_refresh_rate => 250
    )
    port map
    (
      clk_i           => clk_20_i,
      rst_n_i         => rst_20_n,
      led_intensity_i => "1111111",
      led_state_i     => bicolor_led_state_i,
      column_o        => bicolor_led_col_o,
      line_o          => bicolor_led_line_o,
      line_oen_o      => bicolor_led_line_oen_o
    );

end architecture arch;
--==============================================================================
--  architecture end
--==============================================================================
